; DMA VRAM manager

.p816   ; 65816 processor
.a16
.i16
.smart

.export DMAPrepare, DMATryAdd, DMATryAddStrip4, DMAMakeStrip4, DMADoTransfer
.export dmaTableSrc, dmaTableSrcBank, dmaTableDest, dmaTableSize, dmaTableIndex, dmaTotalSize

DMA_OVERHEAD = $30 ; Just a guess currently
DMA_TOTAL_SIZE_LIMIT = $1300 ; Also just a guess

DMA_TABLE_SIZE = 96

.segment "LORAM"

dmaTableSrc: .res DMA_TABLE_SIZE
dmaTableSrcBank: .res DMA_TABLE_SIZE
dmaTableDest: .res DMA_TABLE_SIZE
dmaTableSize: .res DMA_TABLE_SIZE
dmaTableIndex: .res 2
dmaTotalSize: .res 2

.segment "CODE0HI"


; ============================================================================
; DMAPrepare - Called at the beginning of each frame to prepare for any entries
;              added on this frame.
; ---------------------------------------------------------------------------
; Expects anything
; ============================================================================
DMAPrepare:
    php
    rep #$30
    
    stz dmaTableIndex
    stz dmaTotalSize
    
    plp
    rts


; ============================================================================
; DMATryAdd - Checks size of the desired item (stored in A). If it doesn't
;             exceed the DMA limit, the item may be added and the table index
;             is incremented. If it exceeds the limit, the overflow bit is
;             set.
; ---------------------------------------------------------------------------
; Expects any, clobbers A, X, overflow bit
; ============================================================================
DMATryAdd:
    php
    
    rep #$30
    ; Catch size zero (size 65536)
    beq _DontDMA
    
    ldx dmaTableIndex
    cpx #DMA_TABLE_SIZE - 2
    bcs _DontDMA
    
    sta dmaTableSize, x
    clc
    adc dmaTotalSize
    cmp #DMA_TOTAL_SIZE_LIMIT
    bcs _DontDMA
    
    adc #DMA_OVERHEAD
    sta dmaTotalSize
    txa
    inc a
    inc a
    sta dmaTableIndex
    
    plp
    clv
    rts
    
_DontDMA:
    
    plp
    sep #$40 ; Overflow flag indicates we hit DMA limit
    rts


; ============================================================================
; DMATryAddStrip4 - Checks size of the desired item (stored in A) as if it
;             were the size of one strip of four strips total. If it doesn't
;             exceed the DMA limit, the strips may be added and the table
;             index is incremented. If it exceeds the limit, the overflow bit
;             is set. Afterward, DMAMakeStrip4 should be called to copy the
;             information from the first strip to fill in the others.
; ---------------------------------------------------------------------------
; Expects any, clobbers A, X, overflow bit
; ============================================================================
DMATryAddStrip4:
    php
    
    rep #$30
    ; Catch size zero (size 65536)
    beq _DontDMA
    
    ldx dmaTableIndex
    cpx #DMA_TABLE_SIZE - 8
    bcs _DontDMA
    
    sta dmaTableSize, x
    clc
    adc dmaTotalSize
    adc #DMA_OVERHEAD
    asl
    asl
    cmp #DMA_TOTAL_SIZE_LIMIT
    bcs _DontDMA
    
    sta dmaTotalSize
    txa
    clc
    adc #$08
    sta dmaTableIndex
    
    plp
    clv
    rts


; ============================================================================
; DMAMakeStrip4 - Should be called after writing information to the first
;                 strip to copy it to the remaining strips.
; ---------------------------------------------------------------------------
; Expects anything, clobbers A
; ============================================================================
DMAMakeStrip4:
    php
    
    rep #$30
    
    lda dmaTableSrcBank, x
    sta dmaTableSrcBank + 2, x
    sta dmaTableSrcBank + 4, x
    sta dmaTableSrcBank + 6, x
    
    lda dmaTableSize, x
    sta dmaTableSize + 2, x
    sta dmaTableSize + 4, x
    sta dmaTableSize + 6, x
    
    lda dmaTableSrc, x
    clc
    adc dmaTableSize, x
    sta dmaTableSrc + 2, x
    adc dmaTableSize, x
    sta dmaTableSrc + 4, x
    adc dmaTableSize, x
    sta dmaTableSrc + 6, x
    
    lda dmaTableDest, x
    clc
    adc #$0100
    sta dmaTableDest + 2, x
    adc #$0100
    sta dmaTableDest + 4, x
    adc #$0100
    sta dmaTableDest + 6, x
    
    plp
    rts


; ============================================================================
; Expects anything but chops XY, clobbers A, X, Y
; ============================================================================
DMADoTransfer:
    php
    
    rep #$20
    sep #$10
    
    phd
    lda #$4300 ; For faster DMA register accesses
    tcd
    
    ldx dmaTableIndex
    beq _FinishDMATransfer
    
    ;     da ifttt DMA transfer mode
    ldy #%00000001
    sty $00
    ldy #$18    ; VRAM write as destination
    sty $01
    ldy #$80
    sty $2115   ; VMAIN normal, word access
    
_DMATransferLoop:
    
    lda dmaTableSrc - 2, x
    sta $02
    ldy dmaTableSrcBank - 2, x
    sty $04
    lda dmaTableDest - 2, x
    sta $2116
    lda dmaTableSize - 2, x
    sta $05
    ldy #$01
    sty $420b
    
    dex
    dex
    bne _DMATransferLoop
    
_FinishDMATransfer:
    
    pld
    plp
    rts
